/*_##########################################################################
  _##
  _##  Copyright (C) 2012-2014  Kaito Yamada
  _##
  _##########################################################################
*/

package org.pcap4j.packet;

import static org.pcap4j.util.ByteArrays.*;
import java.util.ArrayList;
import java.util.List;
import org.pcap4j.packet.factory.PacketFactories;
import org.pcap4j.packet.namednumber.EtherType;
import org.pcap4j.util.ByteArrays;

/**
 * @author Kaito Yamada
 * @since pcap4j 0.9.11
 */
public final class Dot1qVlanTagPacket extends AbstractPacket {

  /**
   *
   */
  private static final long serialVersionUID = 1522789079803339400L;

  private final Dot1qVlanTagHeader header;
  private final Packet payload;

  /**
   *
   * @param rawData
   * @return a new Dot1qVlanTagPacket object.
   * @throws IllegalRawDataException
   * @throws NullPointerException if the rawData argument is null.
   * @throws IllegalArgumentException if the rawData argument is empty.
   */
  public static Dot1qVlanTagPacket newPacket(byte[] rawData) throws IllegalRawDataException{
    if (rawData == null) {
      throw new NullPointerException("rawData must not be null.");
    }
    if (rawData.length == 0) {
      throw new IllegalArgumentException("rawData is empty.");
    }
    return new Dot1qVlanTagPacket(rawData);
  }

  private Dot1qVlanTagPacket(byte[] rawData) throws IllegalRawDataException {
    this.header = new Dot1qVlanTagHeader(rawData);

    int payloadLength = rawData.length - header.length();
    if (payloadLength > 0) {
      byte[] rawPayload
        = ByteArrays.getSubArray(
            rawData,
            header.length(),
            payloadLength
          );
      this.payload
        = PacketFactories.getFactory(Packet.class, EtherType.class)
            .newInstance(rawPayload, header.getType());
    }
    else {
      this.payload = null;
    }
  }

  private Dot1qVlanTagPacket(Builder builder) {
    if (
         builder == null
      || builder.type == null
    ) {
      StringBuilder sb = new StringBuilder();
      sb.append("builder: ").append(builder)
        .append(" builder.type: ").append(builder.type);
      throw new NullPointerException(sb.toString());
    }

    this.payload = builder.payloadBuilder != null ? builder.payloadBuilder.build() : null;
    this.header = new Dot1qVlanTagHeader(builder);
  }

  @Override
  public Dot1qVlanTagHeader getHeader() {
    return header;
  }

  @Override
  public Packet getPayload() {
    return payload;
  }

  @Override
  public Builder getBuilder() {
    return new Builder(this);
  }

  /**
   * @author Kaito Yamada
   * @since pcap4j 0.9.11
   */
  public static final class Builder extends AbstractBuilder {

    private byte priority;
    private boolean cfi;
    private short vid;
    private EtherType type;
    private Packet.Builder payloadBuilder;

    /**
     *
     */
    public Builder() {}

    private Builder(Dot1qVlanTagPacket packet) {
      this.priority = packet.header.priority;
      this.cfi = packet.header.cfi;
      this.vid = packet.header.vid;
      this.type = packet.header.type;
      this.payloadBuilder = packet.payload != null ? packet.payload.getBuilder() : null;
    }

    /**
     *
     * @param priority
     * @return this Builder object for method chaining.
     */
    public Builder priority(byte priority) {
      this.priority = priority;
      return this;
    }

    /**
     * true: 1, false: 0
     *
     * @param cfi
     * @return this Builder object for method chaining.
     */
    public Builder cfi(boolean cfi) {
      this.cfi = cfi;
      return this;
    }

    /**
     *
     * @param vid
     * @return this Builder object for method chaining.
     */
    public Builder vid(short vid) {
      this.vid = vid;
      return this;
    }

    /**
     *
     * @param type
     * @return this Builder object for method chaining.
     */
    public Builder type(EtherType type) {
      this.type = type;
      return this;
    }

    @Override
    public Builder payloadBuilder(Packet.Builder payloadBuilder) {
      this.payloadBuilder = payloadBuilder;
      return this;
    }

    @Override
    public Packet.Builder getPayloadBuilder() {
      return payloadBuilder;
    }

    @Override
    public Dot1qVlanTagPacket build() {
      return new Dot1qVlanTagPacket(this);
    }

  }

  /**
   * @author Kaito Yamada
   * @since pcap4j 0.9.11
   */
  public static final class Dot1qVlanTagHeader extends AbstractHeader {

    /*
     *   0                                                          15
     * +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     * |Priority   |CFI|                   VID                         |
     * +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     * |                            Type                               |
     * +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     */

    /**
     *
     */
    private static final long serialVersionUID = 7130569411806479522L;

    private static final int PRIORITY_AND_CFI_AND_VID_OFFSET
      = 0;
    private static final int PRIORITY_AND_CFI_AND_VID_SIZE
      = SHORT_SIZE_IN_BYTES;
    private static final int TYPE_OFFSET
      = PRIORITY_AND_CFI_AND_VID_OFFSET + PRIORITY_AND_CFI_AND_VID_SIZE;
    private static final int TYPE_SIZE
      = SHORT_SIZE_IN_BYTES;
    private static final int DOT1Q_TAG_HEADER_SIZE
      = TYPE_OFFSET + TYPE_SIZE;

    private final byte priority;
    private final boolean cfi;
    private final short vid;
    private final EtherType type;

    private Dot1qVlanTagHeader(byte[] rawData) throws IllegalRawDataException {
      if (rawData.length < DOT1Q_TAG_HEADER_SIZE) {
        StringBuilder sb = new StringBuilder(110);
        sb.append("The data is too short to build an IEEE802.1Q Tag header(")
          .append(DOT1Q_TAG_HEADER_SIZE)
          .append(" bytes). data: ")
          .append(ByteArrays.toHexString(rawData, " "));
        throw new IllegalRawDataException(sb.toString());
      }

      short priorityAndCfiAndVid
        = ByteArrays.getShort(rawData, PRIORITY_AND_CFI_AND_VID_OFFSET);

      this.priority = (byte)((priorityAndCfiAndVid & 0xE000) >> 13);
      this.cfi = ((priorityAndCfiAndVid & 0x1000) >> 12) == 1;
      this.vid = (short)(priorityAndCfiAndVid & 0x0FFF);
      this.type
        = EtherType.getInstance(ByteArrays.getShort(rawData, TYPE_OFFSET));
    }

    private Dot1qVlanTagHeader(Builder builder) {
      if ((builder.priority & 0xF8) != 0) {
        throw new IllegalArgumentException(
                "invalid priority: " + builder.priority
              );
      }
      if ((builder.vid & 0xF000) != 0) {
        throw new IllegalArgumentException("invalid vid: " + builder.vid);
      }

      this.priority = builder.priority;
      this.cfi = builder.cfi;
      this.vid = builder.vid;
      this.type = builder.type;
    }

    /**
     *
     * @return priority
     */
    public byte getPriority() {
      return priority;
    }

    /**
     * true: 1, false: 0
     *
     * @return cfi
     */
    public boolean getCfi() {
      return cfi;
    }

    /**
     *
     * @return vid
     */
    public short getVid() {
      return vid;
    }

    /**
     *
     * @return vid
     */
    public int getVidAsInt() {
      return 0x0FFF & vid;
    }

    /**
     *
     * @return type
     */
    public EtherType getType() {
      return type;
    }

    @Override
    protected List<byte[]> getRawFields() {
      List<byte[]> rawFields = new ArrayList<byte[]>();
      rawFields.add(
        ByteArrays.toByteArray(
          (short)((priority << 13) | ((cfi ? 1 : 0) << 12) | vid)
        )
      );
      rawFields.add(ByteArrays.toByteArray(type.value()));
      return rawFields;
    }

    @Override
    public int length() {
      return DOT1Q_TAG_HEADER_SIZE;
    }

    @Override
    protected String buildString() {
      StringBuilder sb = new StringBuilder();
      String ls = System.getProperty("line.separator");

      sb.append("[IEEE802.1Q Tag header (")
        .append(length())
        .append(" bytes)]")
        .append(ls);
      sb.append("  Priority: ")
        .append(priority)
        .append(ls);
      sb.append("  CFI: ")
        .append(cfi ? 1 : 0)
        .append(ls);
      sb.append("  VID: ")
        .append(getVidAsInt())
        .append(ls);
      sb.append("  Type: ")
        .append(type)
        .append(ls);

      return sb.toString();
    }

  }

}
